/*!
  * @license MIT (https://github.com/geotrev/omdomdom/blob/master/LICENSE)
  * Omdomdom v0.3.1 (https://github.com/geotrev/omdomdom)
  * Copyright 2022 George Treviranus <geowtrev@gmail.com>
  */
(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.Omdomdom = {}));
})(this, (function (exports) { 'use strict';

  const DATA_KEY_ATTRIBUTE$1 = "data-key";
  const hasProperty = (obj, prop) => Object.prototype.hasOwnProperty.call(obj, prop);
  const keyIsValid = (map, key) => {
    if (!key) return false;
    if (hasProperty(map, key)) {
      console.warn("[OmDomDom]: Children with duplicate keys detected. Children with duplicate keys will be skipped, resulting in dropped node references. Keys must be unique and non-indexed.");
      return false;
    }
    return true;
  };
  const forEach = (items, fn) => {
    const length = items.length;
    let idx = -1;
    if (!length) return;
    while (++idx < length) {
      if (fn(items[idx], idx) === false) break;
    }
  };
  const createKeyMap = children => {
    const map = {};
    forEach(children, child => {
      const key = child.attributes[DATA_KEY_ATTRIBUTE$1];
      if (keyIsValid(map, key)) {
        map[key] = child;
      }
    });
    for (const _ in map) {
      return map;
    }
  };
  const insertBefore = (parent, child, refNode) => {
    return parent.node.insertBefore(child.node, refNode);
  };
  const assignVNode = (template, vNode) => {
    for (let property in template) {
      vNode[property] = template[property];
    }
  };

  const toHTML = htmlString => {
    const processedDOMString = htmlString.trim().replace(/\s+</g, "<").replace(/>\s+/g, ">");
    const parser = new DOMParser();
    const context = parser.parseFromString(processedDOMString, "text/html");
    return context.body;
  };

  const Types = {
    STRING: "string",
    INT: "number",
    BOOL: "boolean"
  };
  const DomProperties = {};
  const createRecord = (attrName, propName, type) => ({
    attrName,
    propName,
    type
  });
  const stringProps = [["style", "cssText"], ["class", "className"]];
  forEach(stringProps, _ref => {
    let [attr, property] = _ref;
    DomProperties[attr] = createRecord(attr, property || attr, Types.STRING);
  });
  const booleanProps = ["autofocus", "draggable", "hidden", "checked", "multiple", "muted", "selected"];
  forEach(booleanProps, attr => {
    DomProperties[attr] = createRecord(attr, attr, Types.BOOL);
  });
  const integerProps = [["tabindex", "tabIndex"]];
  forEach(integerProps, _ref2 => {
    let [attr, property] = _ref2;
    DomProperties[attr] = createRecord(attr, property, Types.INT);
  });
  const Namespace = {
    xlink: {
      prefix: "xlink:",
      resource: "http://www.w3.org/1999/xlink"
    },
    xml: {
      prefix: "xml:",
      resource: "http://www.w3.org/XML/1998/namespace"
    }
  };

  const setProperty = (element, type, prop, value) => {
    switch (type) {
      case Types.STRING:
        {
          if (prop === DomProperties.style.propName) {
            if (value === null) {
              element.style[prop] = "";
            } else {
              element.style[prop] = value;
            }
          } else if (value === null) {
            element[prop] = "";
          } else {
            element[prop] = value;
          }
          break;
        }
      case Types.INT:
        {
          if (value === null) {
            const attr = prop.toLowerCase();
            element.removeAttribute(attr);
          } else if (value === "0") {
            element[prop] = 0;
          } else if (value === "-1") {
            element[prop] = -1;
          } else {
            const parsed = parseInt(value, 10);
            if (!isNaN(parsed)) {
              element[prop] = parsed;
            }
          }
          break;
        }
      case Types.BOOL:
        {
          if (["", "true"].indexOf(value) < 0) {
            element[prop] = false;
          } else {
            element[prop] = true;
          }
          break;
        }
    }
  };

  const removeAttributes = (vNode, attributes) => {
    forEach(attributes, attrName => {
      if (hasProperty(DomProperties, attrName)) {
        const propertyRecord = DomProperties[attrName];
        setProperty(vNode.node, propertyRecord.type, propertyRecord.propName, null);
      } else {
        if (attrName in vNode.node) {
          setProperty(vNode.node, Types.STRING, attrName, null);
        }
        vNode.node.removeAttribute(attrName);
      }
      delete vNode.attributes[attrName];
    });
  };
  const setAttributes = (vNode, attributes) => {
    for (let attrName in attributes) {
      const value = attributes[attrName];
      vNode.attributes[attrName] = value;
      if (hasProperty(DomProperties, attrName)) {
        const propertyRecord = DomProperties[attrName];
        setProperty(vNode.node, propertyRecord.type, propertyRecord.propName, value);
        continue;
      }
      if (attrName.startsWith(Namespace.xlink.prefix)) {
        vNode.node.setAttributeNS(Namespace.xlink.resource, attrName, value);
        continue;
      }
      if (attrName.startsWith(Namespace.xml.prefix)) {
        vNode.node.setAttributeNS(Namespace.xml.resource, attrName, value);
        continue;
      }
      if (attrName in vNode.node) {
        setProperty(vNode.node, Types.STRING, attrName, value);
      }
      vNode.node.setAttribute(attrName, value || "");
    }
  };
  const getPropertyValues = (element, attributes) => {
    for (let attrName in DomProperties) {
      const propertyRecord = DomProperties[attrName];
      const propName = propertyRecord.propName;
      const attrValue = element.getAttribute(attrName);
      if (attrName === DomProperties.style.attrName) {
        attributes[attrName] = element.style[propName];
      } else if (typeof attrValue === "string") {
        attributes[attrName] = attrValue;
      }
    }
  };
  const getBaseAttributes = element => {
    return Array.prototype.reduce.call(element.attributes, (attributes, attrName) => {
      if (!hasProperty(DomProperties, attrName.name)) {
        attributes[attrName.name] = attrName.value;
      }
      return attributes;
    }, {});
  };
  const getAttributes = element => {
    const attributes = getBaseAttributes(element);
    getPropertyValues(element, attributes);
    return attributes;
  };
  const updateAttributes = (template, vNode) => {
    let removedAttributes = [];
    let changedAttributes = {};
    for (let attrName in vNode.attributes) {
      const oldValue = vNode.attributes[attrName];
      const nextValue = template.attributes[attrName];
      if (oldValue === nextValue) continue;
      if (typeof nextValue === "undefined") {
        removedAttributes.push(attrName);
      }
    }
    for (let attrName in template.attributes) {
      const oldValue = vNode.attributes[attrName];
      const nextValue = template.attributes[attrName];
      if (oldValue === nextValue) continue;
      if (typeof nextValue !== "undefined") {
        changedAttributes[attrName] = nextValue;
      }
    }
    removeAttributes(vNode, removedAttributes);
    setAttributes(vNode, changedAttributes);
  };

  const DATA_KEY_ATTRIBUTE = "data-key";
  function patchChildren(template, vNode, patch) {
    const templateChildrenLength = template.children.length;
    const vNodeChildrenLength = vNode.children.length;
    if (!templateChildrenLength && !vNodeChildrenLength) return;
    const vNodeKeyMap = createKeyMap(vNode.children);
    let nextChildren = Array(templateChildrenLength);
    if (vNodeKeyMap !== undefined) {
      forEach(template.children, (templateChild, idx) => {
        const childNodes = vNode.node.childNodes;
        const key = templateChild.attributes[DATA_KEY_ATTRIBUTE];
        if (Object.prototype.hasOwnProperty.call(vNodeKeyMap, key)) {
          const keyedChild = vNodeKeyMap[key];
          if (Array.prototype.indexOf.call(childNodes, keyedChild.node) !== idx) {
            insertBefore(vNode, keyedChild, childNodes[idx]);
          }
          nextChildren[idx] = keyedChild;
          delete vNodeKeyMap[key];
          patch(templateChild, nextChildren[idx]);
        } else {
          insertBefore(vNode, templateChild, childNodes[idx]);
          nextChildren[idx] = templateChild;
        }
      });
    } else {
      forEach(template.children, (templateChild, idx) => {
        const vNodeChild = vNode.children[idx];
        if (typeof vNodeChild !== "undefined") {
          patch(templateChild, vNodeChild);
          nextChildren[idx] = vNodeChild;
        } else {
          vNode.node.appendChild(templateChild.node);
          nextChildren[idx] = templateChild;
        }
      });
    }
    vNode.children = nextChildren;
    let childNodesLength = vNode.node.childNodes.length;
    let delta = childNodesLength - templateChildrenLength;
    if (delta > 0) {
      while (delta > 0) {
        vNode.node.removeChild(vNode.node.childNodes[childNodesLength - 1]);
        childNodesLength--;
        delta--;
      }
    }
  }

  const patch = (template, vNode, rootNode) => {
    if (!template || !vNode) return;
    rootNode = rootNode || vNode.node.parentNode;
    const contentChanged = template.content && template.content !== vNode.content;
    if (template.type !== vNode.type || contentChanged) {
      rootNode.replaceChild(template.node, vNode.node);
      return assignVNode(template, vNode);
    }
    updateAttributes(template, vNode);
    patchChildren(template, vNode, patch);
  };
  const render = (vNode, root) => {
    root.appendChild(vNode.node);
  };
  const create = function (node) {
    let isSVGContext = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
    if (typeof node === "string") {
      node = toHTML(node);
    }
    const isRoot = node.tagName === "BODY";
    const childNodes = node.childNodes;
    const numChildNodes = childNodes ? childNodes.length : 0;
    if (isRoot) {
      if (numChildNodes > 1) {
        throw new Error("[OmDomDom]: Your element should not have more than one root node.");
      } else if (numChildNodes === 0) {
        throw new Error("[OmDomDom]: Your element should have at least one root node.");
      } else {
        return create(childNodes[0]);
      }
    }
    const type = node.nodeType === 3 ? "text" : node.nodeType === 8 ? "comment" : node.tagName.toLowerCase();
    const isSVG = isSVGContext || type === "svg";
    const attributes = node.nodeType === 1 ? getAttributes(node) : {};
    const content = numChildNodes > 0 ? null : node.textContent;
    const children = Array(numChildNodes);
    forEach(childNodes, (child, idx) => {
      children[idx] = create(child, isSVG);
    });
    return {
      type,
      attributes,
      children,
      content,
      node,
      isSVGContext: isSVG
    };
  };

  exports.create = create;
  exports.patch = patch;
  exports.render = render;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=omdomdom.js.map
